
"""opensignals

Usage:
  opensignals [-v...]
  opensignals -h | --help
  opensignals --version
  opensignals [options] download --dir=<dir> [--recreate]

Options:
  -h --help         Show this screen.
  --version         Show version.
  --verbose=<level> Increase verbosity. [default: 1]

Commands:
    opensignals download --dir=somedirectory [--recreate]
        Download new data.

"""
import logging
from pathlib import Path
from typing import List, Optional

from docopt import docopt

from opensignals import __version__
from opensignals.data import yahoo
from opensignals.features import RSI

def main(argv: Optional[List[str]] = None) -> None:
    """Main entry point.

    Args:
        argv (list of str, optional): Anything passed here will be treated as sys.argv[1:]
            (command-line arguments minus the entry point itself). Useful for testing.

    """
    args = docopt(__doc__, version=__version__)

    if args['--verbose'] and int(args['--verbose']) > 1:
        logging.basicConfig(level=logging.DEBUG)
    elif args['--verbose'] and int(args['--verbose']) == 0:
        logging.basicConfig(level=logging.WARNING)
    else:
        logging.basicConfig(level=logging.INFO)

    if args['download']:
        yahoo.download_data(Path(args['--dir']), args['--recreate'])


if __name__ == '__main__':
    main()



USER_AGENTS = [
    (
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko)'
        ' Chrome/39.0.2171.95 Safari/537.36'
    ),
    (
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko)'
        ' Chrome/42.0.2311.135 Safari/537.36 Edge/12.246'
    )
]


import shutil
import numpy as np
import pandas as pd
import logging
import time as _time
import requests
import random

from tqdm import tqdm
from concurrent import futures
from datetime import datetime, date, time
from dateutil.relativedelta import relativedelta, FR

from opensignals import utils

logger = logging.getLogger(__name__)

AWS_BASE_URL='https://numerai-signals-public-data.s3-us-west-2.amazonaws.com'
SIGNALS_UNIVERSE=f'{AWS_BASE_URL}/latest_universe.csv'
SIGNALS_TICKER_MAP=f'{AWS_BASE_URL}/signals_ticker_map_w_bbg.csv'
SIGNALS_TARGETS=f'{AWS_BASE_URL}/signals_train_val_bbg.csv'

def get_tickers():
    ticker_map = pd.read_csv(SIGNALS_TICKER_MAP)
    ticker_map = ticker_map.dropna(subset=['yahoo'])
    logger.info(f'Number of eligible tickers: {ticker_map.shape[0]}')

    if ticker_map['yahoo'].duplicated().any():
        raise Exception(
            f'Found duplicated {ticker_map["yahoo"].duplicated().values().sum()}'
            ' yahoo tickers'
        )

    if ticker_map['bloomberg_ticker'].duplicated().any():
        raise Exception(
            f'Found duplicated {ticker_map["bloomberg_ticker"].duplicated().values().sum()}'
            ' bloomberg_ticker tickers'
        )

    return ticker_map


def get_ticker_data(db_dir):
    ticker_data = pd.DataFrame({
        'bloomberg_ticker' : pd.Series([], dtype='str'),
        'date' : pd.Series([], dtype='datetime64[ns]')
    })
    if len(list(db_dir.rglob('*.parquet'))) > 0:
        ticker_data = pd.read_parquet(db_dir)

    logger.info(f'Retrieving data for {ticker_data.bloomberg_ticker.unique().shape[0]} '
                 'tickers from the database')

    return ticker_data


def get_ticker_missing(
    ticker_data, ticker_map, last_friday = datetime.today() - relativedelta(weekday=FR(-1))
):
    tickers_available_data = ticker_data.groupby('bloomberg_ticker').agg({'date': [max, min]})
    tickers_available_data.columns = ['date_max', 'date_min']

    eligible_tickers_available_data = ticker_map.merge(
        tickers_available_data.reset_index(),
        on='bloomberg_ticker',
        how='left'
    )

    ticker_not_found = eligible_tickers_available_data.loc[
        eligible_tickers_available_data.date_max.isna(), ['bloomberg_ticker', 'yahoo']
    ]

    ticker_not_found['start'] = '2002-12-01'

    last_friday_52 = last_friday - relativedelta(weeks=52)
    tickers_outdated = eligible_tickers_available_data.loc[
        (
            (eligible_tickers_available_data.date_max < last_friday.strftime('%Y-%m-%d')) &
            (eligible_tickers_available_data.date_max > last_friday_52.strftime('%Y-%m-%d'))
        ),
        ['bloomberg_ticker', 'yahoo', 'date_max']
    ]

    tickers_outdated['start'] = (
        tickers_outdated['date_max'] + pd.DateOffset(1)
    ).dt.strftime('%Y-%m-%d')
    tickers_outdated.drop(columns=['date_max'], inplace=True)

    return pd.concat(
        [ticker_not_found, tickers_outdated]
    )


def get_data(
    db_dir,
    features_generators = [],
    last_friday = datetime.today() - relativedelta(weekday=FR(-1)),
    target='target'
):
    ticker_data = get_ticker_data(db_dir)

    ticker_universe = pd.read_csv(SIGNALS_UNIVERSE)
    ticker_data = ticker_data[ticker_data.bloomberg_ticker.isin(ticker_universe['bloomberg_ticker'])]

    targets = pd.read_csv(SIGNALS_TARGETS)
    targets['date'] = pd.to_datetime(
        targets['friday_date'],
        format='%Y%m%d'
    )
    targets['target_6d'] = targets['target']
    targets['target'] = targets[target]

    feature_names = []
    for features_generator in features_generators:
        ticker_data, feature_names_aux = features_generator.generate_features(ticker_data)
        feature_names.extend(feature_names_aux)

    # merge our feature data with Numerai targets
    ml_data = pd.merge(
        ticker_data, targets,
        on=['date', 'bloomberg_ticker'],
        how='left'
    )

    logger.info(f'Found {ml_data.target.isna().sum()} rows without target, filling with 0.5')
    ml_data['target'] = ml_data['target'].fillna(0.5)

    # convert date to datetime and index on it
    ml_data = ml_data.set_index('date')

    # for training and testing we want clean, complete data only
    ml_data = ml_data.dropna(subset=feature_names)
    # ensure we have only fridays
    ml_data = ml_data[ml_data.index.weekday == 4]
    # drop eras with under 50 observations per era
    ml_data = ml_data[ml_data.index.value_counts() > 50]

    # train test split
    train_data = ml_data[ml_data['data_type'] == 'train']
    test_data = ml_data[ml_data['data_type'] == 'validation']

    # generate live data
    date_string = last_friday.strftime('%Y-%m-%d')
    live_data = ticker_data[ticker_data.date == date_string].copy()

    # get data from the day before, for markets that were closed
    last_thursday = last_friday - relativedelta(days=1)
    thursday_date_string = last_thursday.strftime('%Y-%m-%d')
    thursday_data = ticker_data[ticker_data.date == thursday_date_string]

    # Only select tickers than aren't already present in live_data
    thursday_data = thursday_data[
        ~thursday_data.bloomberg_ticker.isin(live_data.bloomberg_ticker.values)
    ].copy()

    live_data = pd.concat([live_data, thursday_data])
    live_data = live_data.set_index('date')

    return train_data, test_data, live_data, feature_names


def download_tickers(tickers, start):
    start_epoch = int(datetime.strptime(start, '%Y-%m-%d').timestamp())
    end_epoch = int(datetime.combine(date.today(), time()).timestamp())

    pbar = tqdm(
        total=len(tickers),
        unit='tickers'
    )

    dfs = {}
    with futures.ThreadPoolExecutor() as executor:
        _futures = []
        for ticker in tickers:
            _futures.append(
                executor.submit(download_ticker, ticker=ticker, start_epoch=start_epoch, end_epoch=end_epoch)
            )

        for future in futures.as_completed(_futures):
            pbar.update(1)
            ticker, data = future.result()
            dfs[ticker] = data

    pbar.close()

    return pd.concat(dfs)


def download_ticker(ticker, start_epoch, end_epoch):
    def empty_df():
        return pd.DataFrame(columns=[
            "date", "bloomberg_ticker",
            "open", "high", "low", "close",
            "adj_close", "adj_close", "volume",
            "currency", "provider"])

    retries = 20
    backoff = 1
    url = f'https://query2.finance.yahoo.com/v8/finance/chart/{ticker}'
    user_agent = random.choice(utils.USER_AGENTS)
    params = dict(
        period1=start_epoch,
        period2=end_epoch,
        interval='1d',
        events='div,splits',
    )
    data = requests.get(url=url, params=params)
    while(retries > 0):
        retries -= 1
        try:
            data = requests.get(
                url=url,
                params=params,
                headers={'User-Agent': user_agent}
            )
            data_json = data.json()
            quotes = data_json["chart"]["result"][0]
            if "timestamp" not in quotes:
                return ticker, empty_df()

            timestamps = quotes["timestamp"]
            ohlc = quotes["indicators"]["quote"][0]
            volumes = ohlc["volume"]
            opens = ohlc["open"]
            closes = ohlc["close"]
            lows = ohlc["low"]
            highs = ohlc["high"]

            adjclose = closes
            if "adjclose" in quotes["indicators"]:
                adjclose = quotes["indicators"]["adjclose"][0]["adjclose"]

            df = pd.DataFrame({
                "date": pd.to_datetime(timestamps, unit="s").normalize(),
                "bloomberg_ticker": ticker,
                "open": np.array(opens, dtype='float32'),
                "high": np.array(highs, dtype='float32'),
                "low": np.array(lows, dtype='float32'),
                "close": np.array(closes, dtype='float32'),
                "adj_close": np.array(adjclose, dtype='float32'),
                "volume": np.array(volumes, dtype='float32'),
                "currency": quotes['meta']['currency'],
                "provider": 'yahoo'
            })

            return ticker, df.drop_duplicates().dropna()

        except Exception as e:
            _time.sleep(backoff)
            backoff = min(backoff * 2, 30)

    return ticker, empty_df()


def download_data(db_dir, recreate = False):
    if recreate:
        logging.warn(f'Removing dataset {db_dir} to recreate it')
        shutil.rmtree(db_dir, ignore_errors=True)

    db_dir.mkdir(exist_ok=True)

    ticker_data = get_ticker_data(db_dir)
    ticker_map = get_tickers()
    ticker_missing = get_ticker_missing(ticker_data, ticker_map)

    n_ticker_missing = ticker_missing.shape[0]
    if n_ticker_missing <= 0:
        logger.info(f'Dataset up to date')
        return

    logger.info(f'Downloading missing data for {n_ticker_missing} tickers')

    ticker_missing_grouped = ticker_missing.groupby('start').apply(
        lambda x: ' '.join(x.yahoo.astype(str))
    )
    concat_dfs = []
    for start_date, tickers in ticker_missing_grouped.iteritems():
        temp_df = download_tickers(tickers.split(' '), start=start_date)
        if temp_df.empty:
            continue

        temp_df['created_at'] = datetime.now()
        temp_df['volume'] = temp_df['volume'].astype('float64')
        temp_df['bloomberg_ticker'] = temp_df['bloomberg_ticker'].map(
            dict(zip(ticker_map['yahoo'], ticker_map['bloomberg_ticker'])))

        # Yahoo Finance returning previous day in some situations (e.g. Friday in TelAviv markets)
        temp_df = temp_df[temp_df.date >= start_date]

        concat_dfs.append(temp_df)

    if len(concat_dfs) == 0:
        logger.info(f'Dataset up to date')
        return

    df = pd.concat(concat_dfs)
    n_ticker_data = df.bloomberg_ticker.unique().shape[0]
    if n_ticker_data <= 0:
        logger.info(f'Dataset up to date')
        return

    logger.info(f'Storing data for {n_ticker_data} tickers')
    df.to_parquet(db_dir / f'{datetime.utcnow().timestamp()}.parquet', index=False)



"""Contain information about the project."""
__package_name__ = 'opensignals'
__version__ = '0.0.1-rc7'

__description__ = 'Open Signals'

__author__ = 'Jordi Villar'
__author_email__ = 'jrdi.villar@gmail.com'

__license__ = 'Apache License'



import logging

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


class VarChange:
    def __init__(self, num_days=1, variable='adj_close'):
        self.num_days=num_days
        self.variable=variable

    def generate_features(self, ticker_data):
        logger.info(f'generating var change {self.num_days} for {self.variable}...')
        feature_prefix_name = f'{self.variable}_x{self.num_days}'
        ticker_groups = ticker_data.groupby('bloomberg_ticker')
        ticker_data[feature_prefix_name] = ticker_groups[self.variable].transform(
            lambda x: x.shift(self.num_days)
        )

        ticker_data[f'{feature_prefix_name}_diff'] = ticker_data[self.variable] / ticker_data[feature_prefix_name] - 1
        return ticker_data, []


class RSI:
    def __init__(self, num_days=5, interval=10, variable='adj_close'):
        self.num_days=num_days
        self.interval=interval
        self.variable=variable

    @staticmethod
    def relative_strength_index(prices, interval):
        '''
        Computes Relative Strength Index given a price series and lookback interval
        See more here https://www.investopedia.com/terms/r/rsi.asp
        '''
        delta = prices.diff()

        # copy deltas, set losses to 0, get rolling avg
        gains = delta.copy()
        gains[gains < 0] = 0
        avg_gain = gains.rolling(interval).mean()

        # copy deltas, set gains to 0, get rolling avg
        losses = delta.copy()
        losses[losses > 0] = 0
        avg_loss = losses.rolling(interval).mean().abs()

        # calculate relative strength and it's index
        rs = avg_gain / avg_loss
        rsi = 100.0 - (100.0 / (1.0 + rs))
        return rsi

    @staticmethod
    def get_feature_names(num_days, prefix_name):
        # define column names of features, target, and prediction
        feat_quintile_lag = [f'{prefix_name}_quintile_lag_{num}' for num in range(num_days + 1)]
        feat_rsi_diff = [f'{prefix_name}_diff_{num}' for num in range(num_days)]
        feat_rsi_diff_abs = [f'{prefix_name}_abs_diff_{num}' for num in range(num_days)]
        return feat_quintile_lag, feat_rsi_diff, feat_rsi_diff_abs

    def generate_features(self, ticker_data):
        # add Relative Strength Index
        logger.info(f'generating RSI {self.interval} for {self.variable}...')
        feature_prefix_name = f'RSI_{self.interval}_{self.variable}'
        ticker_groups = ticker_data.groupby('bloomberg_ticker')
        ticker_data[feature_prefix_name] = ticker_groups[self.variable].transform(
            lambda x: self.relative_strength_index(x, self.interval)
        )

        # group by era (date)
        logger.debug('grouping by dates...')
        date_groups = ticker_data.groupby('date')

        # create quintile labels within each era, useful for learning relative ranking
        logger.debug('generating RSI quintiles...')
        ticker_data[f'{feature_prefix_name}_quintile'] = date_groups[feature_prefix_name].transform(
            lambda group: pd.qcut(group, 5, labels=False, duplicates='drop')
        )
        ticker_data.dropna(inplace=True)

        (
            feat_quintile_lag, feat_rsi_diff, feat_rsi_diff_abs
        ) = self.get_feature_names(self.num_days, feature_prefix_name)

        # create lagged features grouped by ticker
        logger.debug('grouping by ticker...')
        ticker_groups = ticker_data.groupby('bloomberg_ticker')

        # lag 0 is that day's value, lag 1 is yesterday's value, etc
        logger.debug('generating lagged RSI quintiles...')
        for day in range(self.num_days + 1):
            ticker_data[feat_quintile_lag[day]] = ticker_groups[f'{feature_prefix_name}_quintile'].transform(
                lambda group: group.shift(day)
            )

        # create difference of the lagged features and
        # absolute difference of the lagged features (change in RSI quintile by day)
        logger.debug('generating lagged RSI diffs...')
        for day in range(self.num_days):
            ticker_data[feat_rsi_diff[day]] = (
                ticker_data[feat_quintile_lag[day]] - ticker_data[feat_quintile_lag[day + 1]]
            )
            ticker_data[feat_rsi_diff_abs[day]] = np.abs(ticker_data[feat_rsi_diff[day]])

        return ticker_data, feat_quintile_lag + feat_rsi_diff + feat_rsi_diff_abs


class SMA:
    def __init__(self,  num_days=5, interval=10, variable='adj_close'):
        self.num_days=num_days
        self.interval=interval
        self.variable=variable

    @staticmethod
    def simple_moving_average(prices, interval):
        return prices.rolling(interval).mean()

    @staticmethod
    def get_feature_names(num_days, prefix_name):
        # define column names of features, target, and prediction
        feat_quintile_lag = [f'{prefix_name}_quintile_lag_{num}' for num in range(num_days + 1)]
        feat_rsi_diff = [f'{prefix_name}_diff_{num}' for num in range(num_days)]
        feat_rsi_diff_abs = [f'{prefix_name}_abs_diff_{num}' for num in range(num_days)]
        return feat_quintile_lag, feat_rsi_diff, feat_rsi_diff_abs

    def generate_features(self, ticker_data):
        # add Relative Strength Index
        logger.info(f'generating SMA {self.interval} for {self.variable}...')
        feature_prefix_name = f'SMA_{self.interval}_{self.variable}'
        ticker_groups = ticker_data.groupby('bloomberg_ticker')
        ticker_data[feature_prefix_name] = ticker_groups[self.variable].transform(
            lambda x: self.simple_moving_average(x, self.interval)
        )

        # group by era (date)
        logger.debug('grouping by dates...')
        date_groups = ticker_data.groupby('date')

        # create quintile labels within each era, useful for learning relative ranking
        logger.debug('generating SMA quintiles...')
        ticker_data[f'{feature_prefix_name}_quintile'] = date_groups[feature_prefix_name].transform(
            lambda group: pd.qcut(group, 5, labels=False, duplicates='drop')
        )
        ticker_data.dropna(inplace=True)

        (
            feat_quintile_lag, feat_sma_diff, feat_sma_diff_abs
        ) = self.get_feature_names(self.num_days, feature_prefix_name)

        # create lagged features grouped by ticker
        logger.debug('grouping by ticker...')
        ticker_groups = ticker_data.groupby('bloomberg_ticker')

        # lag 0 is that day's value, lag 1 is yesterday's value, etc
        logger.debug('generating lagged SMA quintiles...')
        for day in range(self.num_days + 1):
            ticker_data[feat_quintile_lag[day]] = ticker_groups[f'{feature_prefix_name}_quintile'].transform(
                lambda group: group.shift(day)
            )

        # create difference of the lagged features and
        # absolute difference of the lagged features (change in SMA quintile by day)
        logger.debug('generating lagged SMA diffs...')
        for day in range(self.num_days):
            ticker_data[feat_sma_diff[day]] = (
                ticker_data[feat_quintile_lag[day]] - ticker_data[feat_quintile_lag[day + 1]]
            )
            ticker_data[feat_sma_diff_abs[day]] = np.abs(ticker_data[feat_sma_diff[day]])

        return ticker_data, feat_quintile_lag + feat_sma_diff + feat_sma_diff_abs



from pathlib import Path
from setuptools import setup, find_packages

metadata_path = Path(__file__).parent / 'src' / 'opensignals' / '__about__.py'
metadata = {}
with metadata_path.open() as file:
    raw_code = file.read()
exec(raw_code, metadata)
metadata = {key.strip('_'): value for key, value in metadata.items()}
metadata['name'] = metadata.pop('package_name')

setup(
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/councilofelders/opensignals',
    project_urls={
        'Bug Tracker': 'https://github.com/councilofelders/opensignals/issues',
    },
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[path.stem for path in Path('src').glob('*.py')],
    python_requires='>=3.7',
    zip_safe=True,
    include_package_data=True,
    install_requires=[
        'docopt',
        'pandas==1.2.2',
        'numpy==1.20.1',
        'pyarrow==3.0.0',
        'requests',
        'tqdm',
    ],
    extras_require=dict(
        test=[
            'pytest==5.1.2',
            'pytest-cov==2.7.1',
            'pytest-flake8==1.0.6',
            'pytest-mypy==0.4.0',
            'pydocstyle==4.0.1',
            'pep8-naming==0.8.1',
            'pytest-docstyle==2.0.0',
            'flake8 == 3.8.1',
        ],
    ),
    entry_points={
        'console_scripts': [
            'opensignals = opensignals.__main__:main',
        ]
    },
    **metadata,
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Development Status :: 3 - Alpha',
        # 'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Topic :: Office/Business :: Financial',
        'Topic :: Office/Business :: Financial :: Investment',
        'Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3',
    ],
)
